import os
import zipfile

from enum import Enum
from time import sleep

import boto3
from appdirs import user_data_dir
from botocore.errorfactory import ClientError

from rich.console import Console

from efemarai.problem_type import ProblemType

console = Console()


class DatasetFormat(Enum):
    """
    Possible dataset formats.
    """

    COCO = "COCO"
    ImageRegression = "ImageRegression"
    ImageNet = "ImageNet"

    @classmethod
    def has(cls, value):
        try:
            cls(value)
            return True
        except ValueError:
            return False


class Dataset:
    """
    Provides dataset related functionality.
    It can be created through the :class:`efemarai.project.Project.create_dataset` method.

    Example:

    .. code-block:: python
        :emphasize-lines: 2,5

        import efemarai as ef
        dataset = ef.Session().project("Name").create_dataset(...)
        # do something else
        dataset.reload()
        if dataset.loaded:
            print('Dataset has finished loading')

    Example (2):

    .. code-block:: python

        import efemarai as ef
        project = ef.Session().project("Name")
        dataset = project.dataset("Dataset Name")
        print(f"Classes: {dataset.classes}")

    """

    @staticmethod
    def create(
        project,
        name,
        format,
        stage,
        data_url,
        annotations_url,
        credentials,
        upload,
        num_datapoints,
        mask_generation,
        min_asset_size,
    ):
        """
        Create a dataset. A more convenient way is to use `project.create_model(...)`.
        """
        if name is None or data_url is None:
            return None

        if format is not None:
            assert format in ("COCO", "ImageNet", "ImageRegression")
        else:
            format = "COCO"

        if stage is not None:
            assert stage in ("train", "validation", "test")
        else:
            stage = "test"

        assert mask_generation in (None, "Simple", "Advanced")

        session = project._session

        response = session._put(
            f"api/dataset/undefined/{project.id}",
            json={
                "name": name,
                "stage": stage,
                "data_url": data_url,
                "annotations_url": annotations_url,
                "access_token": credentials,
                "upload": upload,
                "projectId": project.id,
            },
        )
        dataset_id = response["id"]

        if upload:
            endpoint = f"api/dataset/{dataset_id}/upload"
            if annotations_url is not None:
                session._upload(annotations_url, endpoint)
            session._upload(data_url, endpoint)
            session._post(
                endpoint,
                json={
                    "num_samples": num_datapoints,
                    "mask_generation": mask_generation,
                    "min_asset_size": min_asset_size,
                },
            )

        return Dataset(
            project,
            dataset_id,
            name,
            format,
            stage,
            data_url,
            annotations_url,
            False,
            [],
        )

    def __init__(
        self,
        project,
        id,
        name,
        format,
        stage,
        data_url,
        annotations_url,
        loaded,
        classes,
    ):
        self.project = (
            project  #: (:class:`efemarai.project.Project`) Associated project.
        )
        self.id = id
        self.name = name  #: (str) Name of the dataset.
        self.format = format  #: (str) Format of the dataset.
        self.stage = stage  #: (str) Stage of the dataset.
        self.data_url = data_url
        self.annotations_url = annotations_url
        self.loaded = loaded  #: (bool) Is the dataset loaded.
        self.classes = classes  #: (list [str]) List of names of the classes. `None` is used for those ids that don't have a corresponding name.

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  format={self.format}"
        res += f"\n  stage={self.stage}"
        res += f"\n  data_url={self.data_url}"
        res += f"\n  annotations_url={self.annotations_url}"
        res += f"\n  loaded={self.loaded}"
        res += f"\n  classes={self.classes}"
        res += f"\n)"
        return res

    @staticmethod
    def _parse_classes(dataset_classes):
        if dataset_classes == []:
            return []

        max_index = max([c["index"] for c in dataset_classes]) + 1

        classes = [None] * max_index
        for c in dataset_classes:
            classes[c["index"]] = c["name"]

        return classes

    def delete(self, delete_dependants=False):
        """
        Delete the dataset. You cannot delete an object that is used in a stress test or a baseline (delete those first). This cannot be undone.
        """

        self.project._session._delete(
            f"api/dataset/{self.id}/{self.project.id}/{delete_dependants}"
        )

    def reload(self):
        """
        Reloads the dataset *in place* from the remote endpoint and return it.

        Returns:
            The updated dataset object.
        """

        endpoint = f"api/dataset/{self.id}"
        dataset_details = self.project._session._get(endpoint)

        self.name = dataset_details["name"]
        self.format = dataset_details["format"]
        self.stage = dataset_details["stage"]
        self.data_url = dataset_details["data_url"]
        self.annotations_url = dataset_details["annotations_url"]
        self.loaded = dataset_details["loaded"]
        self.classes = self._parse_classes(dataset_details["classes"])

        return self

    def download(
        self,
        num_samples=None,
        dataset_format=None,
        path=None,
        unzip=True,
        ignore_cache=False,
    ):
        """
        Download the dataset locally.

        Args:
            num_samples (int): Number of samples to download. Leave `None` for all.
            dataset_format (str): What format to download the dataset. Currently supported include `COCO`, `YOLO`, `VOC`, `ImageNet`, `CSV`.
            path (str): The path where to download the data.
            unzip (bool): Should the downloaded zip be unzipped.
            ignore_cache (bool): Force regeneration of the dataset by ignoring the cache. May lead to slower subsequent calls.
        """

        if not self.loaded:
            console.print(":poop: Dataset has not finished loading.", style="red")
            return None

        if path is None:
            path = user_data_dir(appname="efemarai")

        path = os.path.join(path, self.id)

        if dataset_format is None:
            if self.project.problem_type == ProblemType.Classification:
                dataset_format = "imagenet"
            elif self.project.problem_type == ProblemType.ObjectDetection:
                dataset_format = "coco"
            elif self.project.problem_type == ProblemType.InstanceSegmentation:
                dataset_format = "coco"

        if dataset_format is None:
            console.print(":poop: Unsupported problem type.", style="red")
            return None

        if not ignore_cache:
            name = os.path.join(path, f"dataset_{dataset_format}")

            if num_samples:
                name += f"_{num_samples}"

            if os.path.exists(name):
                return name

            name += ".zip"
            if os.path.exists(name + ".zip"):
                return name

        access = self.project._session._post(
            "api/downloadDataset",
            json={
                "id": self.id,
                "format": dataset_format,
                "num_samples": num_samples,
                "async_download": True,
            },
        )

        s3 = boto3.client(
            "s3",
            aws_access_key_id=access["AccessKeyId"],
            aws_secret_access_key=access["SecretAccessKey"],
            aws_session_token=access["SessionToken"],
            endpoint_url=access["Url"],
        )

        with console.status(f"Preparing '{self.name}' dataset download"):
            while True:
                try:
                    response = s3.head_object(
                        Bucket=access["Bucket"], Key=access["ObjectKey"]
                    )
                    size = response["ContentLength"]
                    break
                except ClientError:
                    sleep(1)

        with self.project._session._progress_bar() as progress:
            task = progress.add_task(f"Downloading dataset ", total=float(size))
            callback = lambda num_bytes: progress.advance(task, num_bytes)

            os.makedirs(path, exist_ok=True)
            filename = os.path.join(path, os.path.basename(access["ObjectKey"]))

            s3.download_file(
                access["Bucket"], access["ObjectKey"], filename, Callback=callback
            )

        if unzip:
            with console.status(f"Unzipping dataset"):
                dirname = os.path.splitext(filename)[0]
                with zipfile.ZipFile(filename, "r") as f:
                    f.extractall(dirname)

                os.remove(filename)

                filename = dirname

        console.print(
            (f":heavy_check_mark: Downloaded '{self.name}' dataset to\n  {filename}"),
            style="green",
        )

        return filename



import os
import tempfile

from rich.console import Console

console = Console()


class ModelRepository:
    def __init__(self, url, branch=None, hash=None, access_token=None, check=False):
        self.url = url
        self.branch = branch if branch is not None else "main"
        self.hash = hash
        self.access_token = access_token

        if check:
            self._check()

    def _check(self):
        class Progress(RemoteProgress):
            def update(self, op_code, cur_count, max_count=None, message=""):
                print(f"Git clone: {self._cur_line}")

        with tempfile.TemporaryDirectory() as dir_name:
            try:
                url = furl(self.url)
                url.username = self.access_token
                url.password = "x-oauth-basic"
                repo = Repo.clone_from(
                    url.tostr(),
                    dir_name,
                    branch=self.branch,
                    depth=1,
                    single_branch=True,
                )
            except git.exc.GitError as e:
                message = f"Unable to clone repository at '{self.url}'"
                console.print(f":poop: {message}", style="red")
                raise ValueError(message)

            try:
                if self.hash is None:
                    self.hash = repo.head.commit.hexsha

                repo.commit(self.hash)
            except git.exc.GitError as e:
                message = f"Commit '{self.hash}' does not exist."
                console.print(f":poop: {message}", style="red")
                raise ValueError(message)

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n      url={self.url}"
        res += f"\n      branch={self.branch}"
        res += "\n    )"
        return res


class ModelFile:
    def __init__(self, name, url, upload=False, credentials=None, check=False):
        self.name = name
        self.url = url
        self.upload = upload
        self.credentials = credentials

        if check:
            self._check()

    def _check(self):
        if self.upload and not os.path.exists(self.url):
            message = f"Path '{self.url}' does not exist."
            console.print(f":poop: {message}", style="red")
            raise ValueError(message)

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n      name={self.name}"
        res += f"\n      url={self.url}"
        res += f"\n      upload={self.upload}"
        res += f"\n    )"
        return res


class Model:
    """
    Provides model related functionality.

    It should be created through the :class:`efemarai.project.Project.create_model` method.

    Example:

    .. code-block:: python
        :emphasize-lines: 2

        import efemarai as ef
        ef.Session().project("Name").create_model(...)
    """

    @staticmethod
    def create(project, name, description, version, repository, files, check=False):
        """
        Creates a model.

        You should use :func:`project.create_model` instead.
        """
        if name is None:
            raise ValueError("Missing model name")

        if repository is None:
            raise ValueError("Missing model repository")

        if files is None:
            files = []

        if not isinstance(repository, ModelRepository):
            repository = ModelRepository(**repository, check=check)

        files = [
            ModelFile(**f, check=check) if not isinstance(f, ModelFile) else f
            for f in files
        ]

        session = project._session
        response = session._put(
            f"api/model/undefined/{project.id}",
            json={
                "name": name,
                "description": description,
                "version": version,
                "repository": {
                    "url": repository.url,
                    "branch": repository.branch,
                    "hash": repository.hash,
                    "access_token": repository.access_token,
                },
                "files": [
                    {
                        "name": f.name,
                        "url": f.url,
                        "upload": f.upload,
                        "credentials": f.credentials,
                    }
                    for f in files
                ],
            },
        )
        model_id = response["id"]

        for f in files:
            if f.upload:
                session._upload(f.url, f"api/model/{model_id}/upload")

        return Model(project, model_id, name, description, version, repository, files)

    def __init__(self, project, id, name, description, version, repository, files):
        self.project = project
        self.id = id
        self.name = name
        self.description = description
        self.version = version
        self.repository = repository
        self.files = files

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  description={self.description}"
        res += f"\n  version={self.version}"
        res += f"\n  repository={self.repository}"
        res += f"\n  files={self.files}"
        res += f"\n)"
        return res

    def delete(self, delete_dependants=False):
        """
        Deletes the model.

        You cannot delete an object that is used in a stress test or a baseline
        (delete those first). Deletion cannot be undone.
        """
        self.project._session._delete(
            f"api/model/{self.id}/{self.project.id}/{delete_dependants}"
        )



import os
import time
from pathlib import Path

import click
import click_completion

from rich.console import Console
from rich.prompt import Confirm
from rich.table import Table

import efemarai as ef


click_completion.init()
console = Console()


@click.group()
def main():
    """Efemarai CLI."""
    pass


@main.command()
@click.option("-c", "--config", help="Optional configuration file.")
@click.option(
    "--noinput", is_flag=True, default=False, help="Avoid user input prompts."
)
@click.option("--username", help="Efemarai username.")
@click.option("--url", help="Efemarai URL.")
def init(config, noinput, username, url):
    """Initialize Efemarai."""
    if noinput:
        ef.Session._setup(url, username, password=os.environ.get("EFEMARAI_PASSWORD"))
    else:
        ef.Session._user_setup(config_file=config)


@main.group()
def project():
    """Manage projects."""
    pass


@project.command("create")
@click.argument("definition-file", required=True)
@click.option(
    "--exists-ok/--exists-not-ok",
    default=False,
    help="Append differences if project already exists.",
)
def project_create(definition_file, exists_ok):
    """Create a project following the specified configuration file.

    definition_file (str): YAML file containing project definition."""
    if definition_file == ".":
        definition_file = "efemarai.yaml"

    result = ef.Session().load(definition_file, exists_ok=exists_ok)
    console.print(result)


@project.command("list")
def project_list():
    """Lists the projects associated with the current user."""
    table = Table(box=None)
    [table.add_column(x) for x in ["Id", "Name", "Problem Type"]]
    for m in ef.Session().projects:
        table.add_row(m.id, m.name, str(m.problem_type))
    console.print(table)


@project.command("delete")
@click.argument("project", required=True)
@click.option("-y", "--yes", default=False, is_flag=True, help="Confirm deletion.")
def project_delete(project, yes):
    """Delete the specified project."""
    if project == ".":
        project = "efemarai.yaml"

    project_name = project

    if project.endswith((".yaml", ".yml")):
        project = _get_project(file=project)
    else:
        project = ef.Session().project(project)

    if not project:
        console.print(f"Project '{project_name}' does not exist.")
        return

    if yes or Confirm.ask(
        f"Do you want to delete project [bold]{project.name}[/bold] including all stress tests, models, datasets and domains?",
        default=False,
    ):
        project.delete(delete_dependants=True)


@main.group()
def model():
    """Manage models."""
    pass


@model.command("list")
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def model_list(file):
    """Lists the models in the current project."""
    project = _get_project(file)
    if not project:
        return

    table = Table(box=None)
    [table.add_column(x) for x in ["Id", "Name"]]
    for m in project.models:
        table.add_row(m.id, m.name)
    console.print(table)


@model.command("create")
@click.argument("definition-file", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
@click.option("-v", "--verbose", count=True, help="Print resulting model.")
@click.option(
    "--exists-ok/--exists-not-ok",
    default=False,
    help="Specifies if a model can be overwritten.",
)
def model_create(definition_file, file, verbose, exists_ok):
    """Create a model in the current project."""

    project = _get_project(file)

    definition = ef.Session._read_config(definition_file)

    try:
        model_definitions = definition["models"]
    except KeyError:
        console.print(f":poop: Key 'models' not found in {definition}", style="red")
        exit(1)

    for model_definition in model_definitions:
        try:
            model = project.create_model(**model_definition, exists_ok=exists_ok)
        except ValueError as e:
            console.print(e)
            exit(1)

    if verbose:
        console.print(model)

    console.print(model.id)

    return model


@model.command("delete")
@click.argument("model", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
@click.option(
    "-d",
    "--delete-dependants",
    is_flag=True,
    show_default=True,
    default=False,
    help="Delete all entities that depend on the model",
)
def model_delete(model, file, delete_dependants):
    """Delete a model from the current project.

    MODEL - the name or ID of the model."""
    project = _get_project(file)
    if not project:
        return

    if _check_for_multiple_entities(project.models, model):
        console.print("There are multiple models with the given:\n")
        models = [t for t in project.models if t.name == test]
        _print_table(models)
        console.print(
            f"\nRun the command with a specific model id: [bold green]$ efemarai model delete {models[0].id}",
        )
        return

    model_name = model
    model = project.model(model)
    if not model:
        console.print(f":poop: Model '{model_name}' does not exist.", style="red")
        return
    model.delete(delete_dependants)


@main.group()
def domain():
    """Manage domains."""
    pass


@domain.command("list")
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def domain_list(file):
    """Lists the domains in the current project."""
    project = _get_project(file)
    if not project:
        return

    table = Table(box=None)
    [table.add_column(x) for x in ["Id", "Name"]]
    for d in project.domains:
        table.add_row(d.id, d.name)
    console.print(table)


@domain.command("create")
@click.argument("definition-file", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def domain_create(definition_file, file):
    """Create a domain in the current project.

    definition_file (str): YAML file containing domain definition."""
    project = _get_project(file)
    if not project:
        return

    definition = ef.Session._read_config(definition_file)
    domain = project.create_domain(**definition["domain"])
    return domain


@domain.command("delete")
@click.argument("domain", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
@click.option(
    "-d",
    "--delete-dependants",
    is_flag=True,
    show_default=True,
    default=False,
    help="Delete all entities that depend on the domain",
)
def domain_delete(domain, file, delete_dependants):
    """Delete a domain from the current project.

    DOMAIN - the name or ID of the domain."""
    project = _get_project(file)
    if not project:
        return

    if _check_for_multiple_entities(project.domains, domain):
        console.print("There are multiple domains with the given name:\n")
        domains = [t for t in project.domains if t.name == test]
        _print_table(domains)
        console.print(
            f"\nRun the command with a specific domain id: [bold green]$ efemarai domain delete {domains[0].id}",
        )
        return

    domain_name = domain
    domain = project.domain(domain)
    if not domain:
        console.print(f":poop: Domain '{domain_name}' does not exist.", style="red")
        return
    domain.delete(delete_dependants)


@domain.command("download")
@click.argument("domain", required=True)
@click.option("-o", "--output", default=None, help="Optional domain output file.")
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def domain_download(domain, output, file):
    """Download a domain.

    DOMAIN - the name of the domain."""
    project = _get_project(file)
    if not project:
        return

    domain = project.domain(domain)
    filename = domain.download(filename=output)
    console.print(
        (f":heavy_check_mark: Downloaded '{domain.name}' reports to: \n  {filename}"),
        style="green",
    )


@main.group()
def dataset():
    """Manage datasets."""
    pass


@dataset.command("list")
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def dataset_list(file):
    """Lists the datasets in the current project."""
    project = _get_project(file)
    if not project:
        return

    table = Table(box=None)
    [table.add_column(x) for x in ["Id", "Name", "Loaded"]]
    for d in project.datasets:
        table.add_row(d.id, d.name, str(d.loaded))
    console.print(table)


@dataset.command("create")
@click.argument("definition_file", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def dataset_create(definition_file, file):
    """Create a dataset in the current project.

    definition_file (str): YAML file containing dataset definition."""
    project = _get_project(file)
    if not project:
        return

    definition = ef.Session._read_config(definition_file)
    dataset = project.create_dataset(**definition["dataset"])
    return dataset


@dataset.command("delete")
@click.argument("dataset", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
@click.option(
    "-d",
    "--delete-dependants",
    is_flag=True,
    show_default=True,
    default=False,
    help="Delete all entities that depend on the dataset",
)
def dataset_delete(dataset, file, delete_dependants):
    """Delete a dataset from the current project.

    DATASET - the name or ID of the dataset."""
    project = _get_project(file)
    if not project:
        return

    if _check_for_multiple_entities(project.datasets, dataset):
        console.print("There are multiple datasets with the given name:\n")
        datasets = [t for t in project.datasets if t.name == test]
        _print_table(datasets)
        console.print(
            f"\nRun the command with a specific dataset id: [bold green]$ efemarai dataset delete {datasets[0].id}",
        )
        return

    dataset_name = dataset
    dataset = project.dataset(dataset)
    if not dataset:
        console.print(f":poop: Dataset '{dataset_name}' does not exist.", style="red")
        return
    dataset.delete(delete_dependants)


@main.group()
def test():
    """Manage stress tests."""
    pass


@test.command("list")
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def test_list(file):
    """Lists the stress tests in the current project."""
    project = _get_project(file)
    if not project:
        return

    _print_table(project.stress_tests)


@test.command("run")
@click.argument("definition-file", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
@click.option(
    "-w", "--wait", default=False, is_flag=True, help="Wait for stress test to finish."
)
def test_run(definition_file, file, wait):
    """Run a stress test.

    definition_file (str): YAML file containing stress test definition."""
    project = _get_project(file)
    if not project:
        return

    definition = ef.Session._read_config(definition_file)
    test = project.create_stress_test(**definition["test"])
    cfg = ef.Session._read_config()
    console.print(f"{cfg['url']}project/{project.id}/runs/{test.id}")

    if wait:
        while not test.reload().finished:
            if test.failed:
                exit(1)
            time.sleep(0.1)
        # If not finished normally, an error has occurred
        if not test.finished:
            exit(1)


@test.command("delete")
@click.argument("test", required=True)
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def test_delete(test, file):
    """Delete a stress test from the current project.

    TEST - Name, ID or yaml file of the stress test."""
    project = _get_project(file)
    if not project:
        return

    if _check_for_multiple_entities(project.stress_tests, test):
        console.print("There are multiple stress tests with the given name:\n")
        tests = [t for t in project.stress_tests if t.name == test]
        _print_table(tests)
        console.print(
            f"\nRun the command with a specific stress test id: [bold green]$ efemarai test delete {tests[0].id}",
        )
        return

    test_name = test

    if test.endswith((".yaml", ".yml")):
        test_name = _get_test_name(test)

    test = project.stress_test(test_name)

    if test is None:
        console.print(f":poop: Stress test '{test_name}' does not exist.", style="red")
        exit(1)

    test.delete()


@test.command("download")
@click.argument("test", required=True)
@click.option("--min_score", default=0, help="Minimum score for the samples.")
@click.option("--include_dataset", default=False, help="Include original test dataset.")
@click.option("--path", default=None, help="Path to the downloaded files.")
@click.option("--unzip", default=True, help="Whether to unzip the resulting file.")
@click.option("--ignore_cache", default=False, help="Ignore local cache.")
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def test_download(test, min_score, include_dataset, path, unzip, ignore_cache, file):
    """Download the stress test vulnerabilities dataset.

    TEST - Name, ID or yaml file of the stress test to download."""
    test_name = test
    project = _get_project(file)
    if not project:
        return

    if test.endswith((".yaml", ".yml")):
        test_name = _get_test_name(test)

    test = project.stress_test(test_name)

    if test is None:
        console.print(f":poop: Stress test '{test_name}' does not exist.", style="red")
        exit(1)

    test.vulnerabilities_dataset(
        min_score=min_score,
        include_dataset=include_dataset,
        path=path,
        unzip=unzip,
        ignore_cache=ignore_cache,
    )


@test.command("reports")
@click.argument("test", required=True)
@click.option("-o", "--output", default=None, help="Optional output file.")
@click.option(
    "-f", "--file", help="Name of the Efemarai file (Default is 'efemarai.yaml')"
)
def test_reports(test, output, file):
    """Export the stress test reports.

    TEST - Name, ID or yaml file of the stress test."""
    test_name = test
    project = _get_project(file)
    if not project:
        return

    if test.endswith((".yaml", ".yml")):
        test_name = _get_test_name(test)

    test = project.stress_test(test_name)

    if test is None:
        console.print(f":poop: Stress test '{test_name}' does not exist.", style="red")
        exit(1)

    filename = test.download_reports(filename=output)
    console.print(
        (f":heavy_check_mark: Downloaded '{test.name}' reports to: \n  {filename}"),
        style="green",
    )


def _print_table(tests):
    table = Table(box=None)
    [table.add_column(x) for x in ["Id", "Name", "Model", "Dataset", "Domain", "State"]]
    for t in tests:
        table.add_row(
            t.id, t.name, t.model.name, t.dataset.name, t.domain.name, str(t.state)
        )
    console.print(table)


def _check_for_multiple_entities(entities, name):
    length = len(list(filter(lambda x: x.name == name, entities)))
    return length > 1


def _get_project(file=None):
    if file is None:
        file = "efemarai.yaml"

    if not Path(file).is_file():
        console.print(
            f":poop: Cannot find 'efemarai.yaml' in the current directory.", style="red"
        )
        exit(1)

    conf = ef.Session()._load_config_file(file)
    if "project" not in conf or "name" not in conf["project"]:
        console.print(
            f":poop: '{file}' file not configured properly (does not container project and name within).",
            style="red",
        )
        exit(1)

    name = conf["project"]["name"]
    project = ef.Session().project(name)
    if not project:
        console.print(f"Project '{name}' does not exist.")
        exit(1)

    return project


def _get_test_name(file):
    config = ef.Session()._load_config_file(file)
    return config["test"]["name"]


if __name__ == "__main__":
    main()



from efemarai.dataset import Dataset
from efemarai.domain import Domain
from efemarai.model import Model, ModelFile, ModelRepository
from efemarai.problem_type import ProblemType
from efemarai.stress_test import StressTest


class Project:
    """
    Provides project related functionality.
    A project is the main building block around which an ML model evaluation is performed.

    A project is an association of a set of ML models, datasets and domains.
    """

    @staticmethod
    def create(session, name, description, problem_type, exists_ok=False):
        """
        Creates a project.

        You should use `ef.Session().create_project(...)` instead.
        """
        if name is None or not ProblemType.has(problem_type):
            raise ValueError(
                f"Project name '{name}' is not specified or problem type '{problem_type}' not recognized."
            )

        existing_project = next((p for p in session.projects if p.name == name), None)
        if existing_project:
            if exists_ok:
                return existing_project
            else:
                raise ValueError(f"Project {name} already exists and exists_ok=False.")

        response = session._put(
            "api/project",
            json={
                "name": name,
                "description": description,
                "problem_type": problem_type,
            },
        )
        return Project(session, response["id"], name, description, problem_type)

    def __init__(self, session, id, name, description, problem_type):
        self._session = session
        self.id = id
        self.name = name  #: (str) Name of the project.
        self.description = description  #: (str) Description of the project.
        self.problem_type = ProblemType(
            problem_type
        )  #: (:class:`efemarai.problem_type.ProblemType`) Project problem type.

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  description={self.description}"
        res += f"\n  problem_type={self.problem_type}"
        res += f"\n)"
        return res

    @property
    def models(self):
        """Returns a list of the models associated with the project."""

        return [
            Model(
                project=self,
                id=model["id"],
                name=model["name"],
                description=model.get("description"),
                version=model.get("version"),
                repository=ModelRepository(
                    url=model["repository"]["url"],
                    branch=model["repository"].get("branch"),
                    hash=model["repository"].get("hash"),
                ),
                files=[
                    ModelFile(name=f["name"], url=f["url"], upload=f["upload"])
                    for f in model["files"]
                ],
            )
            for model in self._session._get(f"api/models/{self.id}")
        ]

    def model(self, model) -> Model:
        """Returns the model specified by the name.

        Args:
            model (str): A model identifier. It can either be the name of the model or its id.

        Returns:
            :class:`efemarai.model.Model`: A model object or `None` if it doesn't exist.
        """

        model = next((m for m in self.models if m.name == model or m.id == model), None)
        return model

    def create_model(
        self,
        name,
        description=None,
        version=None,
        repository=None,
        files=None,
        exists_ok=False,
        **kwargs,
    ) -> Model:
        """Creates a model.

        Args:
            name (str): Name of the model.
            description (str, optional): Free text description.
            version (str, optional): Free text description.
            repository (dict): Model repository info. See `ef.model.ModelRepository`.
            files (list[dict]): Files needed to load the model e.g. weights or config
                files. See `ef.model.ModelFile`.
            exists_ok (bool, optional): If the current named model exist, should it be
                returned or an exception raised.

        Raises:
            ValueError: If the model exists and `exists_ok` is set to `False`.

        Returns:
            :class:`efemarai.model.Model`: The created model.
        """

        existing_model = next((m for m in self.models if m.name == name), None)
        if existing_model:
            if exists_ok:
                return existing_model
            else:
                raise ValueError(f"Model {name} already exists and exists_ok=False.")

        model = Model.create(self, name, description, version, repository, files)
        return model

    @property
    def datasets(self):
        """
        Returns a list of the datasets associated with the project.

        Returns:
            list[:class:`efemarai.dataset.Dataset`]: A list of available datasets.
        """

        return [
            Dataset(
                self,
                dataset["id"],
                dataset["name"],
                dataset["format"],
                dataset["stage"],
                dataset["data_url"],
                dataset.get("annotations_url"),
                dataset["loaded"],
                Dataset._parse_classes(dataset["classes"]),
            )
            for dataset in self._session._get(f"api/datasets/{self.id}")
        ]

    def dataset(self, dataset) -> Dataset:
        """
        Returns the dataset specified by the name or id.

        Args:
            dataset (str): A dataset identifier. It can either be the name of the dataset or its id.

        Returns:
            :class:`efemarai.dataset.Dataset`: A dataset object or `None` if it doesn't exist.
        """

        dataset = next(
            (d for d in self.datasets if d.name == dataset or d.id == dataset), None
        )
        return dataset

    def create_dataset(
        self,
        name,
        format,
        stage,
        data_url,
        annotations_url=None,
        credentials=None,
        upload=False,
        num_datapoints=None,
        mask_generation=None,
        exists_ok=False,
        min_asset_size=15,
        **kwargs,
    ) -> Dataset:
        """Creates a dataset.

        Args:
            name (str): A dataset name.
            format (str): A format of the dataset. Possible formats are specified under `ef.dataset.DatasetFormat`.
            state (str): Dataset stage - one of [`training`, `validation`, `test`].
            data_url (str): A url to fetch data. Can be local or remote.
            annotations_url (str, optional): A url that specifies the annotation file. Can be local or remote.
            credentials (str, optional): If the dataset if remote, Specify the `access token` or `{username}:{password}` to access.
            upload (bool, optional): Should the data be uploaded to Efemarai? This would make evaluation and testing much faster.
            num_datapoints (int, optional): Should a subset of the datapoints be loaded as a dataset? Specify the number of dataponts. Useful for quick testing. Use `None` to load all.
            mask_generation (str): What strategy should be used for mask generation.
            exists_ok (bool, optional): If the current named dataset exist, should it be returned or an exception raised.
            min_asset_size (int, optional): Assets with width and hight smaller than this will be ignored. Default value of 15.

        Raises:
            ValueError: If the dataset exists and `exists_ok` is set to `False`.

        Returns:
            :class:`efemarai.dataset.Dataset`: The created dataset.
        """

        existing_dataset = next((d for d in self.datasets if d.name == name), None)
        if existing_dataset:
            if exists_ok:
                return existing_dataset
            else:
                raise ValueError(f"Dataset {name} already exists and exists_ok=False.")

        dataset = Dataset.create(
            self,
            name,
            format,
            stage,
            data_url,
            annotations_url,
            credentials,
            upload,
            num_datapoints,
            mask_generation,
            min_asset_size,
        )

        return dataset

    @property
    def domains(self):
        """
        Returns a list of the domains associated with the project.

        Returns:
            list [:class:`efemarai.domain.Domain`]: A list of available domains.
        """

        return [
            Domain(
                self,
                domain["id"],
                domain["name"],
                domain["transformations"],
                domain["graph"],
            )
            for domain in self._session._get(f"api/domains/{self.id}")
        ]

    def domain(self, domain) -> Domain:
        """
        Returns the domain specified by the name or id.

        Args:
            domain (str): A domain identifier. It can either be the name of the domain or its id.

        Returns:
            :class:`efemarai.domain.Domain`: A domain object or `None` if it doesn't exist.
        """

        domain = next(
            (d for d in self.domains if d.name == domain or d.id == domain), None
        )
        return domain

    def create_domain(
        self, name, transformations, graph, exists_ok=False, **kwargs
    ) -> Domain:
        """
        Creates a domain. **The recommended strategy** is to use the UI to build up the domain visually.
        The domain then can be exported to a yaml configuration file for storage and loaded/reused through `ef.Session().load(...)`.

        Args:
            name (str): The name of the domain.
            transformations (dict): A dictionary of the transformations and their params.
            graph (dict): How the transformations are paired.
            exists_ok (bool, optional): If the current named domain exist, should it be returned or an exception raised.

        Raises:
            ValueError: If the domain exists and `exists_ok` is set to `False`.

        Returns:
            :class:`efemarai.domain.Domain`: The created domain.
        """

        existing_domain = next((d for d in self.domains if d.name == name), None)
        if existing_domain:
            if exists_ok:
                return existing_domain
            else:
                raise ValueError(f"Domain {name} already exists and exists_ok=False.")

        domain = Domain.create(self, name, transformations, graph)
        return domain

    @property
    def stress_tests(self):
        """
        Returns a list of the stress tests associated with the project.

        Returns:
            list [:class:`efemarai.stress_test.StressTest`]: A list of available stress tests.
        """

        return [
            StressTest(
                self,
                test["id"],
                test["name"],
                test["model"]["id"],
                test["domain"]["id"],
                test["dataset"]["id"],
                test["states"][-1]["name"],
                test["reports"],
            )
            for test in self._session._get(f"api/getRuns/{self.id}")["objects"]
        ]

    def stress_test(self, test) -> StressTest:
        """
        Returns the stress test specified by the name or id.

        Args:
            test (str): A stress test identifier. It can either be the name of the stress test or its id.

        Returns:
            :class:`efemarai.stress_test.StressTest`: A stress test object or `None` if it doesn't exist.
        """
        test = next(
            (t for t in self.stress_tests if t.name == test or t.id == test), None
        )
        return test

    def create_stress_test(
        self,
        name,
        model,
        domain,
        dataset,
        num_samples=25,
        num_runs=40,
        concurrent_runs=4,
        **kwargs,
    ) -> StressTest:
        """
        Creates a stress test.

        Args:
            model (str, :class:`efemarai.model.Model`): The model to stress test.
            domain (str, :class:`efemarai.domain.Domain`): The domain from which to generate samples.
            dataset (str, :class:`efemarai.dataset.Dataset`): The dataset with which to generate samples.
            num_samples (int, optional): Sample count to generate.
            num_runs (int, optional): Number of searches to run.
            concurrent_runs (int, optional): Number of concurrent searches.

        Returns:
            :class:`efemarai.stress_test.StressTest`: A stress test object.
        """

        test = StressTest.create(
            self,
            name,
            model,
            domain,
            dataset,
            num_samples,
            num_runs,
            concurrent_runs,
        )

        return test

    def delete(self, delete_dependants=False):
        """
        Deletes a project, including the domains, datasets, models, stress tests. Cannot be undone.
        """

        self._session._delete(f"api/project/{self.id}/{delete_dependants}")



import os
import re
from datetime import datetime
from glob import glob

import boto3
import requests
import validators
import yaml
from appdirs import user_config_dir

from rich.console import Console
from rich.progress import BarColumn, Progress, SpinnerColumn, TimeRemainingColumn
from rich.prompt import Confirm, Prompt

from smart_open import smart_open

from efemarai.project import Project


console = Console()


class Session:
    """
    Session related functionality. A session is the way to interact with the Efemarai endpoint.
    All commands use the active session to communicate and perform the needed actions.
    """

    @staticmethod
    def _fetch_token(url, username, password):
        try:
            response = requests.post(
                url + "auth/sdk-token/",
                json={"username": username, "password": password},
            )

            if not response.ok:
                console.print(":poop: Invalid username or password", style="red")
                return None

            return response.json()["token"]

        except requests.exceptions.ConnectionError:
            console.print(f":poop: Unreachable URL {url}", style="red")
            return None

    @staticmethod
    def _user_setup(config_file=None):
        console.print(":rocket:  [bold]Welcome to Efemarai![/bold]\n", style="#00a9ff")

        existing_config = Session._read_config(config_file)
        if existing_config is None:
            console.print("Let's set up things quickly.")
        else:
            console.print("URL: ", existing_config["url"])
            console.print("User: ", existing_config["username"])
            console.print()
            if not Confirm.ask(
                "Do you want to overwrite existing config?", default=False
            ):
                return

        console.print()

        username = Prompt.ask("Username")
        password = Prompt.ask("Password", password=True)

        while True:
            url = Prompt.ask("Platform URL", default=f"https://{username}.efemarai.com")

            if not re.match(r"^https?://", url):
                url = "https://" + url

            if not url.endswith("/"):
                url += "/"

            if validators.url(url):
                break

            console.print("[prompt.invalid]:poop: Invald URL")

        return Session._setup(url, username, password, config_file)

    @staticmethod
    def _setup(url, username, password, config_file=None):
        token = Session._fetch_token(url, username, password)

        if token is None:
            return None

        config = {"username": username, "url": url, "token": token}

        if config_file is None:
            config_dir = user_config_dir(appname="efemarai")
            os.makedirs(config_dir, exist_ok=True)
            config_file = os.path.join(config_dir, "config.yaml")

        console.print(f":gear: Saving config in {config_file}", style="green")

        with open(config_file, "w") as f:
            yaml.dump(config, f)

        return config

    @staticmethod
    def _read_config(config_file=None):
        if config_file is None:
            config_file = os.path.join(
                user_config_dir(appname="efemarai"), "config.yaml"
            )

        if not os.path.isfile(config_file):
            return None

        return Session._load_config_file(config_file)

    @staticmethod
    def _save_config_file(data, filename):
        with open(filename, "w") as f:
            yaml.dump(data, f)

    @staticmethod
    def _load_config_file(filename):
        with open(filename) as f:
            contents = f.read()

        contents = os.path.expandvars(contents)

        for match in re.findall("\$\{([a-zA-Z]\w*)\}", contents):
            console.print(
                f":poop: Unknown environment variable '{match}' in '{filename}'",
                style="red",
            )

        return yaml.safe_load(contents)

    @staticmethod
    def _progress_bar():
        return Progress(
            SpinnerColumn(style="green"),
            "[progress.description]{task.description}",
            BarColumn(complete_style="default"),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
            transient=True,
        )

    def __init__(self, token=None, url=None, config_file=None):
        self.token = token
        self.url = url

        if self.token is None or self.url is None:
            config = self._read_config(config_file)

            if config is None:
                config = self._user_setup(config_file)

            self.token = config["token"]
            self.url = config["url"]

        self._access_requests = {}

    def load(self, filename, exists_ok=False) -> dict:
        """Load a configuration file for creating projects, models, datasets and/or domains.

        Args:
            filename (str): A configuration file.
            exists_ok (bool, optional): If any of the components defined in the configuration file exist, should they raise an exception or return the already existing element.

        Riases:
            ValueError: If any of the components exist and `exists_ok` is set to `False`.

        Returns:
            dict: A dictionary of the potentially loaded projects, datasets, domains, models.

        Example:

        .. code-block:: python

            import efemarai as ef
            result = ef.Session().load("efemarai.yaml")
            print(result)
        """

        config = self._load_config_file(filename)

        project = None
        if "project" in config:
            project = self.create_project(**config["project"], exists_ok=exists_ok)

        models = [
            project.create_model(**model_config, exists_ok=exists_ok)
            for model_config in config.get("models", [])
        ]

        datasets = [
            project.create_dataset(**dataset_config, exists_ok=exists_ok)
            for dataset_config in config.get("datasets", [])
        ]

        domains = []
        for domain_config in config.get("domains", []):
            if "file" in domain_config:
                domain_config = self._load_config_file(domain_config["file"])["domain"]

            domains.append(project.create_domain(**domain_config, exists_ok=exists_ok))

        return {
            "project": project,
            "datasets": datasets,
            "domains": domains,
            "models": models,
        }

    @property
    def projects(self):
        """
        Returns the list of projects.

        Returns:
            list [:class:`efemarai.project.Project`]: A list of the available projects.
        """

        return [
            Project(
                self,
                project["id"],
                project["name"],
                project.get("description", ""),
                project["problem_type"],
            )
            for project in self._get("api/projects")
        ]

    def project(self, name):
        """
        Returns a project specified by the name.

        Args:
            name (str): A project identifier. It can either be the name of the project or its id.

        Returns:
            :class:`efemarai.project.Project`: A project object or `None` if it doesn't exist.

        Example:

        .. code-block:: python

            import efemarai as ef
            project = ef.Session().project("Name")
            print(project)
        """

        project = next(
            (p for p in self.projects if p.name == name or p.id == name), None
        )
        return project

    def create_project(
        self, name, description=None, problem_type=None, exists_ok=False, **kwargs
    ):
        """
        Creates a project.

        Args:
            name (str): Name of the project.
            description (str, optional): Description of the project.
            problem_type (str, :class:`efemarai.problem_type.ProblemType`): Problem type of the project.
            exists_ok (bool, optional): If the current named project exist, should it be returned or an exception raised.

        Raises:
            ValueError: If the project exists and `exists_ok` is set to `False`.

        Returns:
            :class:`efemarai.project.Project`: A project object.
        """

        project = Project.create(self, name, description, problem_type, exists_ok)
        return project

    def _get(self, endpoint, json=None, params=None):
        return self._make_request(requests.get, endpoint, json, params)

    def _post(self, endpoint, json=None, params=None):
        return self._make_request(requests.post, endpoint, json, params)

    def _put(self, endpoint, json=None, params=None):
        return self._make_request(requests.put, endpoint, json, params)

    def _delete(self, endpoint, json=None, params=None):
        return self._make_request(requests.delete, endpoint, json, params)

    def _make_request(self, method, endpoint, json=None, params=None):
        url = self.url
        if not url.endswith("/") and not endpoint.startswith("/"):
            url += "/"
        url += endpoint

        try:
            response = method(
                url,
                headers={"Authorization": f"Token {self.token}"},
                json=json,
                params=params,
            )
        except requests.exceptions.ConnectionError as e:
            console.print(
                f":poop: It looks you cannot reach us at {self.url}. If you think you should, send us an email.",
                style="RED",
            )
            raise e

        if not response.ok:
            try:
                message = response.json()["message"]
            except:
                message = response.text

            console.print(
                f":poop: [{response.status_code} {response.reason}] {message}",
                style="RED",
            )

            response.raise_for_status()

        try:
            return response.json()
        except:
            return None

    def _upload(self, from_url, endpoint):
        if os.path.isdir(from_url):
            self._upload_directory(from_url, endpoint)
        else:
            self._upload_file(from_url, endpoint)

    def _upload_file(self, from_url, endpoint):
        object_key = os.path.basename(from_url)

        with self._progress_bar() as progress:
            task = progress.add_task(
                f"Uploading '{object_key}' ",
                total=float(os.path.getsize(from_url)),
            )
            callback = lambda num_bytes: progress.advance(task, num_bytes)
            s3, bucket, prefix = self._get_access(endpoint)
            with smart_open(from_url, "rb") as f:
                s3.upload_fileobj(f, bucket, prefix + object_key, Callback=callback)

        console.print(f":heavy_check_mark: Uploaded '{object_key}'", style="green")

    def _upload_directory(self, from_url, endpoint):
        filenames = [
            filename
            for filename in glob(os.path.join(from_url, "**/*"), recursive=True)
            if os.path.isfile(filename)
        ]

        dirname = os.path.basename(os.path.normpath(from_url))

        s3, bucket, prefix = self._get_access(endpoint)

        with self._progress_bar() as progress:
            task = progress.add_task(
                f"Uploading '{dirname}' ", total=float(len(filenames))
            )
            for index, filename in enumerate(filenames):
                object_key = os.path.join(
                    dirname, os.path.relpath(filename, start=from_url)
                )

                with smart_open(filename, "rb") as f:
                    s3.upload_fileobj(f, bucket, prefix + object_key)

                progress.advance(task)

        console.print(f":heavy_check_mark: Uploaded '{from_url}'", style="green")

    def _get_access(self, endpoint):
        access = self._access_requests.get(endpoint, None)

        if access is None or access["Expiration"] <= datetime.now():
            access = self._get(endpoint)
            access["Expiration"] = datetime.strptime(
                access["Expiration"], "%Y-%m-%dT%H:%M:%SZ"
            )
            self._access_requests[endpoint] = access

        s3 = boto3.client(
            "s3",
            aws_access_key_id=access["AccessKeyId"],
            aws_secret_access_key=access["SecretAccessKey"],
            aws_session_token=access["SessionToken"],
            endpoint_url=access["Url"],
        )

        return s3, access["Bucket"], access["Prefix"]



import re
import os
import zipfile
import json

from enum import Enum
from time import sleep

import boto3
from appdirs import user_data_dir
from botocore.errorfactory import ClientError

from rich.console import Console

from efemarai.problem_type import ProblemType

console = Console()


class JobState(Enum):
    """
    Possible Test Run, Baseline and Dataset States.
    """

    NotStarted = "NotStarted"
    Starting = "Starting"
    Evaluating = "Evaluating"
    Testing = "Testing"
    Loading = "Loading"
    DataLoaded = "Data Loaded"
    CalculatingMetadata = "Calculating Metadata"
    GeneratingReport = "Generating Report"
    Finished = "Finished"
    Failed = "Failed"
    Stopped = "Stopped"


class StressTest:
    """
    Provides stress test related functionality.
    It can be created through a :class:`efemarai.project.Project.create_stress_test` method.

    Example:

    .. code-block:: python
        :emphasize-lines: 2,5,7

        import efemarai as ef
        test = ef.Session().project("Name").create_stress_test(...)
        # do something else
        test.reload()
        if test.finished:
            print(f"Report: {test.report}")
            dataset_filename = test.vulnerabilities_dataset(min_score=0.1)

    Example (2):

    .. code-block:: python

        import efemarai as ef
        project = ef.Session().project("Name")
        test = project.stress_test("Test Name")
        test.download_reports()
    """

    @staticmethod
    def create(
        project,
        name,
        model,
        domain,
        dataset,
        num_samples,
        num_runs,
        concurrent_runs,
    ):
        """Create a stress test. A more convenient way is to use :func:`project.create_stress_test`"""

        if isinstance(model, str):
            model = project.model(model)

        if isinstance(dataset, str):
            dataset = project.dataset(dataset)

        if isinstance(domain, str):
            domain = project.domain(domain)

        response = project._session._post(
            "api/runTest",
            json={
                "name": name,
                "model": model.id,
                "dataset": dataset.id,
                "domain": domain.id,
                "project": project.id,
                "samples_per_run": num_samples,
                "runs_count": num_runs,
                "concurrent_runs": concurrent_runs,
            },
        )
        return StressTest(
            project,
            response["id"],
            name,
            model,
            domain,
            dataset,
            "NotStarted",
            {},
        )

    def delete(self):
        """
        Deletes a stress test. This cannot be undone.
        """
        self.project._session._delete("api/deleteRun?id=" + self.id)

    def __init__(self, project, id, name, model, domain, dataset, state, reports):
        self.project = (
            project  #: (:class:`efemarai.project.Project`) Associated project.
        )
        self.id = id
        self.name = name  #: (str) Name of the stress test.

        if isinstance(model, str):
            self._model = None
            self._model_id = model
        else:
            self._model = model
            self._model_id = model.id

        if isinstance(domain, str):
            self._domain = None
            self._domain_id = domain
        else:
            self._domain = domain
            self._domain_id = domain.id

        if isinstance(dataset, str):
            self._dataset = None
            self._dataset_id = dataset
        else:
            self._dataset = dataset
            self._dataset_id = dataset.id

        self.state = JobState(
            state
        )  # (:class:`efemarai.stress_test.JobState`) State of the stress test.
        self._reports = reports

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  model={self.model.name}"
        res += f"\n  domain={self.domain.name}"
        res += f"\n  dataset={self.dataset.name}"
        res += f"\n  state={self.state}"
        res += f"\n  len(reports)={len(self.reports)}"
        res += f"\n)"
        return res

    @property
    def reports(self):
        """Returns the stress test reports. This can be slow as is potentially fetching a large object."""
        if not self._reports:
            response = self.project._session._get(
                "api/getTestRun", params={"testRunId": self.id}
            )
            self._reports = response["reports"]

        return self._reports

    @property
    def model(self):
        """Returns the model associated with the stress test."""

        if self._model is None:
            self._model = next(m for m in self.project.models if m.id == self._model_id)
        return self._model

    @property
    def domain(self):
        """Returns the domain associated with the stress test."""

        if self._domain is None:
            self._domain = next(
                d for d in self.project.domains if d.id == self._domain_id
            )
        return self._domain

    @property
    def dataset(self):
        """Returns the dataset associated with the stress test."""

        if self._dataset is None:
            self._dataset = next(
                d for d in self.project.datasets if d.id == self._dataset_id
            )
        return self._dataset

    @property
    def finished(self):
        """Returns if the stress test has successfully finished.

        :rtype: bool
        """

        return self.state == JobState.Finished

    @property
    def failed(self):
        """Returns if the stress test has failed.

        :rtype: bool
        """

        return self.state == JobState.Failed

    @property
    def running(self):
        """Returns if the stress test is still running - not failed or finished.

        :rtype: bool
        """

        return self.state != JobState.Finished and self.state != JobState.Failed

    def reload(self):
        """
        Reloads the stress test *in place* from the remote endpoint and return it.

        Rerturns:
            The updated stress test object.
        """

        response = self.project._session._get(
            "api/getTestRun", params={"testRunId": self.id}
        )

        self.state = JobState(response["states"][-1]["name"])
        self._reports = response.get("reports", {})

        return self

    def download_reports(self, filename=None):
        """
        Download any generated reports.

        Args:
            filename (str): Specify the filename used to store the report data.

        Returns:
            str: The filename of the downloaded report.
        """
        if filename is None:
            # Remove non-ascii and non-alphanumeric characters
            filename = re.sub(r"[^A-Za-z0-9 ]", r"", self.name)
            # Collapse repeating spaces
            filename = re.sub(r"  +", r" ", filename)
            # Replace spaces with dashes and convert to lowercase
            filename = filename.replace(" ", "_").lower()
            filename += ".json"

        with open(filename, "w") as f:
            f.write(json.dumps(self.reports))

        return filename

    def vulnerabilities_dataset(
        self,
        min_score=0.0,
        include_dataset=False,
        path=None,
        unzip=True,
        ignore_cache=False,
        export_format=None,
    ):
        """
        Returns the vulnerabilities dataset associated with the stress test.

        Args:
            min_score (float, optional): Minimum score to select samples.
            include_dataset (bool, optional): If the dataset used for the stress test should be included.
            path (str, optional): Path to the downloading location.
            unzip (bool, optional): If the zip file should be unzipped.
            ignore_cache (bool, optional): Force regeneration of the dataset by ignoring the cache. May lead to slower subsequent calls.

        Returns:
            str: The filename of the resulting object.
        """

        if not self.finished:
            console.print(
                (
                    ":warning: Cannot export vulnerabilities "
                    "dataset as stress test is still running"
                ),
                style="yellow",
            )
            return None

        if path is None:
            path = user_data_dir(appname="efemarai")

        path = os.path.join(path, self.id)

        if export_format is None:
            if self.project.problem_type == ProblemType.Classification:
                export_format = "imagenet"
            elif self.project.problem_type == ProblemType.ObjectDetection:
                export_format = "coco"
            elif self.project.problem_type == ProblemType.InstanceSegmentation:
                export_format = "coco"

        if export_format is None:
            console.print(":poop: Unsupported problem type.", style="red")
            return None

        if not ignore_cache:
            name = "vulnerabilities_dataset"
            name += f"_{export_format}_{include_dataset}_{min_score:.3f}"
            cache_name = os.path.join(path, name)
            if os.path.exists(cache_name) or os.path.exists(cache_name + ".zip"):
                return cache_name

        access = self.project._session._post(
            "api/exportDataset",
            json={
                "id": self.id,
                "format": export_format,
                "merge": include_dataset,
                "min_score": min_score,
                "async_download": True,
            },
        )

        s3 = boto3.client(
            "s3",
            aws_access_key_id=access["AccessKeyId"],
            aws_secret_access_key=access["SecretAccessKey"],
            aws_session_token=access["SessionToken"],
            endpoint_url=access["Url"],
        )

        with console.status(f"Generating '{self.name}' vulnerabilities dataset"):
            while True:
                try:
                    response = s3.head_object(
                        Bucket=access["Bucket"], Key=access["ObjectKey"]
                    )
                    size = response["ContentLength"]
                    break
                except ClientError:
                    sleep(1)

        with self.project._session._progress_bar() as progress:
            task = progress.add_task(f"Downloading dataset ", total=float(size))
            callback = lambda num_bytes: progress.advance(task, num_bytes)

            os.makedirs(path, exist_ok=True)
            filename = os.path.join(path, os.path.basename(access["ObjectKey"]))

            s3.download_file(
                access["Bucket"], access["ObjectKey"], filename, Callback=callback
            )

        if unzip:
            with console.status(f"Unzipping dataset"):
                dirname = os.path.splitext(filename)[0]
                with zipfile.ZipFile(filename, "r") as f:
                    f.extractall(dirname)

                os.remove(filename)

                filename = dirname

        console.print(
            (
                f":heavy_check_mark: Downloaded '{self.name}' "
                f"vulnerabilities dataset to \n  {filename}"
            ),
            style="green",
        )

        return filename



import re


class Domain:
    """
    Provides domain related functionality. The easiest way to interact with this object is through the UI.
    It can be created through the :class:`efemarai.project.Project.create_domain` method.
    """

    @staticmethod
    def create(project, name, transformations, graph):
        """
        Create a domain. A more convenient way is to use `project.create_domain(...)`.
        """
        if name is None or name is None or transformations is None or graph is None:
            return None

        session = project._session
        response = session._put(
            f"api/domain/undefined/{project.id}",
            json={"name": name, "projectId": project.id},
        )
        domain_id = response["id"]

        response = session._put(
            f"api/domain/{domain_id}/import-flow",
            json={
                "transformations": transformations,
                "graph": graph,
            },
        )

        return Domain(
            project, domain_id, name, response["transformations"], response["graph"]
        )

    def __init__(self, project, id, name, transformations, graph):
        self.project = (
            project  #: (:class:`efemarai.project.Project`) Associated project.
        )
        self.id = id
        self.name = name  #: (str) Name of the domain
        self.transformations = transformations
        self.graph = graph

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  transformations={self.transformations}"
        res += f"\n  graph={self.graph}"
        res += f"\n)"
        return res

    def download(self, filename=None) -> str:
        """
        Downalod the domain specifications to a local file.

        Args:
            filename (str, optional): Specify the filename used to store the domain data.

        Returns:
            (str) The filename of the downloaded domain.
        """
        if filename is None:
            # Remove non-ascii and non-alphanumeric characters
            filename = re.sub(r"[^A-Za-z0-9 ]", r"", self.name)
            # Collapse repeating spaces
            filename = re.sub(r"  +", r" ", filename)
            # Replace spaces with dashes and convert to lowercase
            filename = filename.replace(" ", "_").lower()
            filename += ".yaml"

        response = self.project._session._get(f"api/domain/{self.id}/export")
        with open(filename, "w") as f:
            f.write(response["definition"])

        return filename

    def delete(self, delete_dependants=False):
        """
        Delete the domain. You cannot delete an object that is used in a stress test or a baseline (delete those first). This cannot be undone.
        """
        self.project._session._delete(
            f"api/domain/{self.id}/{self.project.id}/{delete_dependants}"
        )
